/*
 * Copyright (c) 2020 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: LicenseRef-Nordic-5-Clause
 */

/** @file
 *  @brief Peripheral Heart Rate sample
 */
#include <stddef.h>
#include <string.h>
#include <errno.h>
#include <zephyr/kernel.h>
#include <zephyr/types.h>
#include <zephyr/sys/printk.h>

#include <zephyr/bluetooth/bluetooth.h>
#include <zephyr/bluetooth/conn.h>
#include <zephyr/bluetooth/gatt.h>
#include <zephyr/bluetooth/uuid.h>
#include <zephyr/bluetooth/services/bas.h>
#include <zephyr/bluetooth/services/hrs.h>

#include <dk_buttons_and_leds.h>

#define NON_CONNECTABLE_DEVICE_NAME "Nordic Beacon"

#define NON_CONNECTABLE_ADV_IDX 0
#define CONNECTABLE_ADV_IDX_1   1
#define CONNECTABLE_ADV_IDX_2 	2

#define DEVICE_NAME             CONFIG_BT_DEVICE_NAME
#define DEVICE_NAME_LEN         (sizeof(DEVICE_NAME) - 1)

#define RUN_STATUS_LED          DK_LED1
#define CON_STATUS_LED          DK_LED2
#define RUN_LED_BLINK_INTERVAL  1000
#define NOTIFY_INTERVAL         1000

static uint8_t volatile conn_count;

static void start_advertising(struct k_work *work);
static void notify_work_handler(struct k_work *work);

static K_WORK_DEFINE(start_advertising_worker, start_advertising);
static K_WORK_DELAYABLE_DEFINE(notify_work, notify_work_handler);

static struct bt_le_ext_adv *ext_adv[CONFIG_BT_EXT_ADV_MAX_ADV_SET];

static const struct bt_le_adv_param *non_connectable_adv_param =
	BT_LE_ADV_PARAM(BT_LE_ADV_OPT_USE_NAME,
			0x140, /* 200 ms */
			0x190, /* 250 ms */
			NULL);


static const struct bt_data ad_hrs[] = {
	BT_DATA_BYTES(BT_DATA_FLAGS, (BT_LE_AD_GENERAL | BT_LE_AD_NO_BREDR)),
	BT_DATA_BYTES(BT_DATA_UUID16_ALL, BT_UUID_16_ENCODE(BT_UUID_HRS_VAL)),
	BT_DATA(BT_DATA_NAME_COMPLETE, DEVICE_NAME, DEVICE_NAME_LEN)
};

static const struct bt_data ad_bas[] = {
	BT_DATA_BYTES(BT_DATA_FLAGS, (BT_LE_AD_GENERAL | BT_LE_AD_NO_BREDR)),
	BT_DATA_BYTES(BT_DATA_UUID16_ALL, BT_UUID_16_ENCODE(BT_UUID_BAS_VAL)),
	BT_DATA(BT_DATA_NAME_COMPLETE, "Nordic_BAS", DEVICE_NAME_LEN)
};

static const struct bt_data non_connectable_data[] = {
	BT_DATA_BYTES(BT_DATA_FLAGS, BT_LE_AD_NO_BREDR),
	BT_DATA_BYTES(BT_DATA_URI, /* The URI of the https://www.nordicsemi.com website */
		      0x17, /* UTF-8 code point for “https:” */
		      '/', '/', 'w', 'w', 'w', '.',
		      'n', 'o', 'r', 'd', 'i', 'c', 's', 'e', 'm', 'i', '.',
		      'c', 'o', 'm')
};


static void connected(struct bt_conn *conn, uint8_t conn_err)
{
	int err;
	struct bt_conn_info info;
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	if (conn_err) {
		printk("Connection failed (err %d)\n", conn_err);
		return;
	}

	err = bt_conn_get_info(conn, &info);
	if (err) {
		printk("Failed to get connection info (err %d)\n", err);
	} else {
		const struct bt_conn_le_phy_info *phy_info;
		phy_info = info.le.phy;

		printk("Connected: %s, tx_phy %u, rx_phy %u\n",
		       addr, phy_info->tx_phy, phy_info->rx_phy);
	}

	conn_count++;
	if (conn_count < CONFIG_BT_MAX_CONN) {
		k_work_submit(&start_advertising_worker);
	}

	printk("Connection count: %u\n", conn_count);

	dk_set_led_on(CON_STATUS_LED);

	dk_set_led_on(CON_STATUS_LED);
}

static void disconnected(struct bt_conn *conn, uint8_t reason)
{
	printk("Disconnected (reason 0x%02x)\n", reason);

	k_work_submit(&start_advertising_worker);

	dk_set_led_off(CON_STATUS_LED);
}

BT_CONN_CB_DEFINE(conn_callbacks) = {
	.connected = connected,
	.disconnected = disconnected,
};

static int non_connectable_adv_create(void)
{
	int err;

	err = bt_set_name(NON_CONNECTABLE_DEVICE_NAME);
	if (err) {
		printk("Failed to set device name (err %d)\n", err);
		return err;
	}

	err = bt_le_ext_adv_create(non_connectable_adv_param, NULL, &ext_adv[NON_CONNECTABLE_ADV_IDX]);
	if (err) {
		printk("Failed to create non-connectable advertiser set (err %d)\n", err);
		return err;
	}

	err = bt_le_ext_adv_set_data(ext_adv[NON_CONNECTABLE_ADV_IDX], non_connectable_data, ARRAY_SIZE(non_connectable_data), NULL, 0);
	if (err) {
		printk("Failed to set non-connectable advertising data (err %d)\n", err);
		return err;
	}

	err = bt_le_ext_adv_start(ext_adv[NON_CONNECTABLE_ADV_IDX], BT_LE_EXT_ADV_START_DEFAULT);
	if (err) {
		printk("Failed to start non-connectable advertising (err %d)\n", err);
	}

	return err;
}

static int connectable_adv_create(void)
{
	int err;
	struct bt_le_adv_param param =
		BT_LE_ADV_PARAM_INIT(BT_LE_ADV_OPT_CONNECTABLE,
				     BT_GAP_ADV_FAST_INT_MIN_2,
				     BT_GAP_ADV_FAST_INT_MAX_2,
				     NULL);

	
	err = bt_le_ext_adv_create(&param, NULL, &ext_adv[CONNECTABLE_ADV_IDX_1]);
	if (err) {
		printk("Failed to create advertiser set (err %d)\n", err);
		return err;
	}

	printk("Created adv: %p\n", &ext_adv[CONNECTABLE_ADV_IDX_1]);

	err = bt_le_ext_adv_set_data(ext_adv[CONNECTABLE_ADV_IDX_1], ad_hrs, ARRAY_SIZE(ad_hrs), NULL, 0);
	if (err) {
		printk("Failed to set advertising data for CONNECTABLE_ADV_IDX_1 (err %d)\n", err);
		return err;
	}

	err = bt_le_ext_adv_create(&param, NULL, &ext_adv[CONNECTABLE_ADV_IDX_2]);
	if (err) {
		printk("Failed to create advertiser set (err %d)\n", err);
		return err;
	}

	printk("Created adv: %p\n", &ext_adv[CONNECTABLE_ADV_IDX_2]);

	err = bt_le_ext_adv_set_data(ext_adv[CONNECTABLE_ADV_IDX_2], ad_bas, ARRAY_SIZE(ad_bas), NULL, 0);
	if (err) {
		printk("Failed to set advertising data for CONNECTABLE_ADV_IDX_2 (err %d)\n", err);
		return err;
	}

	return 0;
}

static void start_advertising(struct k_work *work)
{
	int err;

	/* TODO: Keep track of which adv. set is active, and only try start 
     * the adv. set that is not active. 
	 */

	err = bt_le_ext_adv_start(ext_adv[CONNECTABLE_ADV_IDX_1], NULL);
	if (err) {
		printk("Failed to start connectable advertising set (err %d)\n", err);
	}
	else {
		printk("Connectable advertising started 'CONNECTABLE_ADV_IDX_1'\n");
	}

	err = bt_le_ext_adv_start(ext_adv[CONNECTABLE_ADV_IDX_2], NULL);
	if (err) {
		printk("Failed to start connectable advertising set (err %d)\n", err);
	}
	else {
		printk("Connectable advertising started 'CONNECTABLE_ADV_IDX_2'\n");
	}

}

static void bas_notify(void)
{
	uint8_t battery_level = bt_bas_get_battery_level();

	__ASSERT_NO_MSG(battery_level > 0);

	battery_level--;

	if (!battery_level) {
		battery_level = 100;
	}

	bt_bas_set_battery_level(battery_level);
}

static void hrs_notify(void)
{
	static uint8_t heartrate = 100;

	heartrate++;
	if (heartrate == 160) {
		heartrate = 100;
	}

	bt_hrs_notify(heartrate);
}

static void notify_work_handler(struct k_work *work)
{
	/* Services data simulation. */
	hrs_notify();
	bas_notify();

	k_work_reschedule(k_work_delayable_from_work(work), K_MSEC(NOTIFY_INTERVAL));
}

void main(void)
{
	uint32_t led_status = 0;
	int err;

	printk("Starting Bluetooth Peripheral HR example\n");

	err = dk_leds_init();
	if (err) {
		printk("LEDs init failed (err %d)\n", err);
		return;
	}

	err = bt_enable(NULL);
	if (err) {
		printk("Bluetooth init failed (err %d)\n", err);
		return;
	}

	printk("Bluetooth initialized\n");

	err = connectable_adv_create();
	if (err) {
		printk("Connectable advertising failed to create (err %d)\n", err);
		return;
	}

	err = non_connectable_adv_create();
	if (err) {
		printk("Non-connectable advertising failed to create (err %d)\n", err);
		return;
	}

	k_work_submit(&start_advertising_worker);
	k_work_schedule(&notify_work, K_NO_WAIT);

	for (;;) {
		dk_set_led(RUN_STATUS_LED, (++led_status) % 2);
		k_sleep(K_MSEC(RUN_LED_BLINK_INTERVAL));
	}
}
